/*  Project: Open Source Gmini Tools
    file: unpack.cpp, main procedure for AOS unpacker

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   To do:
     * replicate bug in original Archos packer for firmware versions
       Gmini120: v1.2 and v1.8,  Gmini220: v1.6, and GminiSP: v1.2
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "types.h"
#include "util.h"
#define UNPACKNAME "unpack"

char *filebase;
u32 printMiniHexDumpNumSpaces;

bool avstyle;
u32 getu32( const u8 * src ) {
   return avstyle ? lu32( src ) : bu32( src );
}

u32 getByteCount( const u8 * data, u32 length ) {
   u8 flag = 0;
   u8 used = 7;
   u32 count = 0;
   const u8 *end = data + length;
   while( data < end ) {
      flag >>= 1;
      if( ++used == 8 ) {
         flag = *( data++ );
         used = 0;
      }
      if( flag & 0x01 )
         ++count;
      else
         count += ( *( ++data ) & 0x0f ) + 3;
      ++data;
   }
   return count;
}

u8 *guessXORkey( const u8 * src, u32 length ) {
   u8 *temp = new u8[length];
   u32 maxzeros = 1;
   u32 keylength = 1;
   for( u32 guess = keylength; guess <= 128; ++guess ) {
      // eliminate keybits
      for( u32 pos = 0; pos < length; ++pos )
         temp[pos] = src[pos] ^ src[( pos + guess ) % length];
      // count zeros
      u32 zeros = 0;
      for( u32 pos = 0; pos < length; ++pos )
         if( temp[pos] == 0 )
            ++zeros;
      if( zeros > maxzeros ) {
         printf( "%d: %d\n", guess, zeros );
         // reasonable checking for multiples of actual length (I think)
         if( ( ( guess % keylength ) != 0 ) || ( zeros > ( maxzeros + ( maxzeros >> 1 ) ) ) ) {
            maxzeros = zeros;
            keylength = guess;
         }
      }
   }
   printf( "probable key length is: %d\n", keylength );
   u8 *key = new u8[keylength + 1];
   key[keylength] = '\0';
   u32 bytecount[256];
   for( u32 i = 0; i < keylength; ++i ) {
      u32 c;
      for( c = 0; c < 256; ++c )
         bytecount[c] = '\0';
      /* gather byte frequency statistics */
      for( c = i; c < length; c += keylength ) {
         u8 v = src[c];
         if( isprint( v ) )
            ++bytecount[v];
      }
      u32 topval = 0;
      u8 topchar = 0;
      for( c = 0; c < 256; ++c )
         if( bytecount[c] > topval ) {
            topval = bytecount[c];
            topchar = c;
         }
      key[i] = topchar;
   }
   printf( "guessed key: %s\n", key );
   return key;
}

void printMiniHexDump( u8 * buf, u32 leng ) {
   u32 i;
   int c;
   char spaces[255];
   u32 TRUNCLENG = 40 - printMiniHexDumpNumSpaces;
   if( leng <= 0 )
      return;
   for( i = 0; i < printMiniHexDumpNumSpaces; i++ )
      spaces[i] = 0x20;
   spaces[i] = 0x0;
   if( leng > TRUNCLENG ) {
      printf( "(dump truncated, printing 0x%x of 0x%x bytes)\n", TRUNCLENG, leng );
      leng = TRUNCLENG;
   }
   printf( spaces );
   for( i = 0; i < leng; i++ )
      printf( "%.02x", buf[i] );
   if( leng < TRUNCLENG )
      printf( "\n" );
   printf( spaces );
   for( i = 0; i < leng; i++ ) {
      c = buf[i] & 0xff;
      printf( " %c", isprint( c ) ? c : '.' );
   }
   if( leng < TRUNCLENG )
      printf( "\n" );
}

void displayGmini( const char *type, u8 * ccod, u32 csize, u32 ucsize ) {
   // AJZ header ID field
   char head[5];
   u8 *image = ccod + 16;
   u32 size;
   cstr( head, ccod, 4 );
   printf( "%s: SIGNATURE: %s\n", type, head );

   // Possibly buggy compressed size stored in AJZ header
   size = lu32( ccod + 8 );
   if( size != csize )
      printf( "%s: BUG in Archos packer: compressed size: 0x%x\n", type, size );
   printf( "%s: Compressed size of firmware: 0x%x\n", type, csize );

   // Checksum stored in AJZ header
   printf( "%s: Checksum: %08x\n", type, lu32( ccod + 12 ) );

   // Checksum of AJZ payload
   u32 sumdata = chksum32( ccod + 16, csize );
   if( lu32( ccod + 12 ) != sumdata )
      printf( "%s: Checksum of data (%08x) doesn't match header\n", type, sumdata );

   // Decrypt firmware Image
   u8 *key = guessXORkey( image, csize );
   XOR( image, csize, key, strlen( ( char * )key ) );

   char *compname = new char[strlen( filebase ) + 6];
   sprintf( compname, "%s.%s", filebase, "comp" );
   printf( "%s: Writing compressed firmware to %s\n", type, compname );
   writeFile( compname, image, csize );
   delete compname;
   // Decompress firmware Image
   // Possibly buggy uncompressed size stored in AJZ header
   size = lu32( ccod + 4 );
   u32 actual = getByteCount( image, csize );
   if( size != ucsize )
      printf( "%s: BUG in Archos packer: uncompressed size.: 0x%x\n", type, size );
   if( size != actual )
      printf( "%s: BUG in Archos packer: uncompressed size..: 0x%x\n", type, size );
   if( ucsize != actual )
      printf( "%s: BUG in Archos packer: uncompressed size...: 0x%x\n", type, ucsize );
   printf( "%s: Uncompressed Size: 0x%x\n", type, actual );
   u8 *firmware = new u8[ucsize];
   inflate( image, csize, firmware, ucsize );

   // Write unencrypted/decompressed file to disk
   char *filename = new char[strlen( filebase ) + 5];
   sprintf( filename, "%s.%s", filebase, "aaz" );
   printf( "%s: Writing firmware to %s\n", type, filename );
   writeFile( filename, firmware, actual );
   delete filename;
   delete firmware;
}

void displayAV( const char *type, u8 * ccod, u32 csize ) {
   printf( "header values: 0x%08x, 0x%08x, 0x%08x\n", getu32( ccod ), getu32( ccod + 4 ), getu32( ccod + 8 ) );
   printf( "don't know how to decrypt this yet, sorry...\n" );
}

void displayChunk( u8 * chunk, u32 leng ) {
   printf( "displayChunk, length = %d\n", leng );
   char type[5] = "null";
   u32 cleng;
   u32 ucsize = 0;
   u8 *end = chunk + leng;
   while( chunk + 8 < end ) {
      cstr( type, chunk, 4 );
      printf( "Chunk type: %s\n", type );
      cleng = getu32( chunk + 4 );
      if( strcmp( type, "RIFF" ) == 0 || strcmp( type, "LIST" ) == 0 ) {
         if( strncmp( ( char * )( chunk + 8 ), "AOS1", 4 ) == 0 ) {
            printf( "AV style AOS file identified, switching to little endian\n" );
            avstyle = true;
            cleng = getu32( chunk + 4 );
            displayChunk( chunk + 12, cleng );
         }
         else
            displayChunk( chunk + 8, cleng );
      }
      else if( strcmp( type, "SWID" ) == 0 ) {
         char vid[5];
         vid[4] = 0;
         strncpy( vid, ( char * )( chunk + 8 ), 4 );
         printf( "%s: %s, %s\n", type, vid, chunk + 12 );
      }
      else if( strcmp( type, "FLSH" ) == 0 ) {
         ucsize = getu32( chunk + 12 );
         printf( "%s: load 0x%x bytes @ 0x%x\n", type, getu32( chunk + 12 ), getu32( chunk + 8 ) );
      }
      else if( strcmp( type, "CCOD" ) == 0 ) {
         if( avstyle )
            displayAV( type, chunk + 8, cleng - 8 );
         else
            displayGmini( type, chunk + 12, getu32( chunk + 8 ) - 16, ucsize );
      }
      else if( strcmp( type, "SIGN" ) == 0 ) {
         char *filename = new char[strlen( filebase ) + 6];
         sprintf( filename, "%s.%s", filebase, "sign" );
         printf( "%s: Writing SIGN chunk to %s\n", type, filename );
         writeFile( filename, chunk + 8, cleng - 8 );
      }
      else
         printMiniHexDump( chunk + 8, ( cleng >= 8 ) ? cleng - 8 : 0 );
      chunk += ( ( cleng == 0 ) ? 8 : cleng );
   }
}

int processAOS( const char *file ) {
   avstyle = false;
   u8 *aos;
   int length;
   if( ( length = loadFile( file, &aos ) ) == -1 )
      return -1;
   printf( "File: %s (0x%x bytes)\n", file, length );
   printMiniHexDumpNumSpaces = 0;
   displayChunk( aos, length );
   delete aos;
   return 1;
}

int main( int argc, char **argv ) {
   int n = 1;
   if( argc < 2 )
      printf( "Usage: %s file.aos [file2.aos ...]\n", UNPACKNAME );
   else
      for( ; n < argc; ++n ) {
         filebase = basename( argv[n] );
         processAOS( argv[n] );
         delete filebase;
      }
   return 0;
}
